@@ -12,39 +12,53 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from __future__ import annotations
+
 import argparse
+import contextlib
 import enum
 import hashlib
 import os
+import pathlib
 import re
 import sys
 import unicodedata
+from types import TracebackType
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
-    Dict,
+    Generator,
     Iterable,
-    List,
     Mapping,
-    Optional,
+    NoReturn,
     Sequence,
-    Tuple,
-    Union,
 )
 
-import py
-
 import nox.command
-from nox import _typing
 from nox._decorators import Func
 from nox.logger import logger
 from nox.virtualenv import CondaEnv, PassthroughEnv, ProcessEnv, VirtualEnv
 
-if _typing.TYPE_CHECKING:
+if TYPE_CHECKING:
     from nox.manifest import Manifest
 
 
-def _normalize_path(envdir: str, path: Union[str, bytes]) -> str:
+@contextlib.contextmanager
+def _chdir(path: str) -> Generator[None, None, None]:
+    """
+    Change the current working directory to the given path.
+    Follows python 3.11's chdir behaviour.
+    """
+    cwd = os.getcwd()
+    try:
+        os.chdir(path)
+        yield
+    finally:
+        os.chdir(cwd)
+
+
+def _normalize_path(envdir: str, path: str | bytes) -> str:
     """Normalizes a string to be a "safe" filesystem path for a virtualenv."""
     if isinstance(path, bytes):
         path = path.decode("utf-8")
@@ -63,15 +77,15 @@ def _normalize_path(envdir: str, path: Union[str, bytes]) -> str:
             logger.warning("The virtualenv name was hashed to avoid being too long.")
         else:
             logger.error(
-                "The virtualenv path {} is too long and will cause issues on "
+                f"The virtualenv path {full_path} is too long and will cause issues on "
                 "some environments. Use the --envdir path to modify where "
-                "nox stores virtualenvs.".format(full_path)
+                "Nox stores virtualenvs."
             )
 
     return full_path
 
 
-def _dblquote_pkg_install_args(args: Tuple[str, ...]) -> Tuple[str, ...]:
+def _dblquote_pkg_install_args(args: tuple[str, ...]) -> tuple[str, ...]:
     """Double-quote package install arguments in case they contain '>' or '<' symbols"""
 
     # routine used to handle a single arg
@@ -79,20 +93,18 @@ def _dblquote_pkg_install_args(args: Tuple[str, ...]) -> Tuple[str, ...]:
         # sanity check: we need an even number of double-quotes
         if pkg_req_str.count('"') % 2 != 0:
             raise ValueError(
-                "ill-formated argument with odd number of quotes: %s" % pkg_req_str
+                f"ill-formatted argument with odd number of quotes: {pkg_req_str}"
             )
 
         if "<" in pkg_req_str or ">" in pkg_req_str:
-            if pkg_req_str[0] == '"' and pkg_req_str[-1] == '"':
+            if pkg_req_str[0] == pkg_req_str[-1] == '"':
                 # already double-quoted string
                 return pkg_req_str
             else:
                 # need to double-quote string
                 if '"' in pkg_req_str:
-                    raise ValueError(
-                        "Cannot escape requirement string: %s" % pkg_req_str
-                    )
-                return '"%s"' % pkg_req_str
+                    raise ValueError(f"Cannot escape requirement string: {pkg_req_str}")
+                return f'"{pkg_req_str}"'
         else:
             # no dangerous char: no need to double-quote string
             return pkg_req_str
@@ -116,6 +128,23 @@ class Status(enum.Enum):
     SKIPPED = 2
 
 
+class _WorkingDirContext:
+    def __init__(self, dir: str | os.PathLike[str]) -> None:
+        self._prev_working_dir = os.getcwd()
+        os.chdir(dir)
+
+    def __enter__(self) -> _WorkingDirContext:
+        return self
+
+    def __exit__(
+        self,
+        exc_type: type[BaseException] | None,
+        exc_value: BaseException | None,
+        traceback: TracebackType | None,
+    ) -> None:
+        os.chdir(self._prev_working_dir)
+
+
 class Session:
     """The Session object is passed into each user-defined session function.
 
@@ -125,11 +154,11 @@ class Session:
 
     __slots__ = ("_runner",)
 
-    def __init__(self, runner: "SessionRunner") -> None:
+    def __init__(self, runner: SessionRunner) -> None:
         self._runner = runner
 
     @property
-    def __dict__(self) -> "Dict[str, SessionRunner]":  # type: ignore
+    def __dict__(self) -> dict[str, SessionRunner]:  # type: ignore[override]
         """Attribute dictionary for object inspection.
 
         This is needed because ``__slots__`` turns off ``__dict__`` by
@@ -144,12 +173,12 @@ class Session:
         return self._runner.friendly_name
 
     @property
-    def env(self) -> dict:
+    def env(self) -> dict[str, str]:
         """A dictionary of environment variables to pass into all commands."""
         return self.virtualenv.env
 
     @property
-    def posargs(self) -> List[str]:
+    def posargs(self) -> list[str]:
         """Any extra arguments from the ``nox`` commandline or :class:`Session.notify`."""
         return self._runner.posargs
 
@@ -162,12 +191,12 @@ class Session:
         return venv
 
     @property
-    def python(self) -> Optional[Union[str, Sequence[str], bool]]:
+    def python(self) -> str | Sequence[str] | bool | None:
         """The python version passed into ``@nox.session``."""
         return self._runner.func.python
 
     @property
-    def bin_paths(self) -> Optional[List[str]]:
+    def bin_paths(self) -> list[str] | None:
         """The bin directories for the virtualenv."""
         return self.virtualenv.bin_paths
 
@@ -183,36 +212,66 @@ class Session:
         """Create, and return, a temporary directory."""
         tmpdir = os.path.join(self._runner.envdir, "tmp")
         os.makedirs(tmpdir, exist_ok=True)
-        self.env["TMPDIR"] = tmpdir
+        self.env["TMPDIR"] = os.path.abspath(tmpdir)
         return tmpdir
 
+    @property
+    def cache_dir(self) -> pathlib.Path:
+        """Create and return a 'shared cache' directory to be used across sessions."""
+        path = pathlib.Path(self._runner.global_config.envdir).joinpath(".cache")
+        path.mkdir(exist_ok=True)
+        return path
+
     @property
     def interactive(self) -> bool:
         """Returns True if Nox is being run in an interactive session or False otherwise."""
         return not self._runner.global_config.non_interactive and sys.stdin.isatty()
 
-    def chdir(self, dir: Union[str, os.PathLike]) -> None:
-        """Change the current working directory."""
-        self.log("cd {}".format(dir))
-        os.chdir(dir)
+    @property
+    def invoked_from(self) -> str:
+        """The directory that Nox was originally invoked from.
+
+        Since you can use the ``--noxfile / -f`` command-line
+        argument to run a Noxfile in a location different from your shell's
+        current working directory, Nox automatically changes the working directory
+        to the Noxfile's directory before running any sessions. This gives
+        you the original working directory that Nox was invoked form.
+        """
+        return self._runner.global_config.invoked_from  # type: ignore[no-any-return]
+
+    def chdir(self, dir: str | os.PathLike[str]) -> _WorkingDirContext:
+        """Change the current working directory.
+
+        Can be used as a context manager to automatically restore the working directory::
+
+            with session.chdir("somewhere/deep/in/monorepo"):
+                # Runs in "/somewhere/deep/in/monorepo"
+                session.run("pytest")
+
+            # Runs in original working directory
+            session.run("flake8")
+
+        """
+        self.log(f"cd {dir}")
+        return _WorkingDirContext(dir)
 
     cd = chdir
     """An alias for :meth:`chdir`."""
 
     def _run_func(
-        self, func: Callable, args: Iterable[Any], kwargs: Mapping[str, Any]
+        self, func: Callable[..., Any], args: Iterable[Any], kwargs: Mapping[str, Any]
     ) -> Any:
         """Legacy support for running a function through :func`run`."""
-        self.log("{}(args={!r}, kwargs={!r})".format(func, args, kwargs))
+        self.log(f"{func}(args={args!r}, kwargs={kwargs!r})")
         try:
             return func(*args, **kwargs)
         except Exception as e:
-            logger.exception("Function {!r} raised {!r}.".format(func, e))
-            raise nox.command.CommandFailed()
+            logger.exception(f"Function {func!r} raised {e!r}.")
+            raise nox.command.CommandFailed() from e
 
     def run(
-        self, *args: str, env: Mapping[str, str] = None, **kwargs: Any
-    ) -> Optional[Any]:
+        self, *args: str, env: Mapping[str, str] | None = None, **kwargs: Any
+    ) -> Any | None:
         """Run a command.
 
         Commands must be specified as a list of strings, for example::
@@ -231,7 +290,18 @@ class Session:
                 'bash', '-c', 'echo $SOME_ENV',
                 env={'SOME_ENV': 'Hello'})
 
-        You can also tell nox to treat non-zero exit codes as success using
+        You can extend the shutdown timeout to allow long-running cleanup tasks to
+        complete before being terminated. For example, if you wanted to allow ``pytest``
+        extra time to clean up large projects in the case that Nox receives an
+        interrupt signal from your build system and needs to terminate its child
+        processes::
+
+            session.run(
+                'pytest', '-k', 'long_cleanup',
+                interrupt_timeout=10.0,
+                terminate_timeout=2.0)
+
+        You can also tell Nox to treat non-zero exit codes as success using
         ``success_codes``. For example, if you wanted to treat the ``pytest``
         "tests discovered, but none selected" error as success::
 
@@ -244,10 +314,31 @@ class Session:
 
             session.run('cmd', '/c', 'del', 'docs/modules.rst')
 
+        If ``session.run`` fails, it will stop the session and will not run the next steps.
+        Basically, this will raise a Python exception. Taking this in count, you can use a
+        ``try...finally`` block for cleanup runs, that will run even if the other runs fail::
+
+           try:
+               session.run("coverage", "run", "-m", "pytest")
+           finally:
+               # Display coverage report even when tests fail.
+               session.run("coverage", "report")
+
+        If you pass ``silent=True``, you can capture the output of a command that would
+        otherwise be shown to the user. For example to get the current Git commit ID::
+
+            out = session.run(
+                "git", "rev-parse", "--short", "HEAD",
+                external=True, silent=True
+            )
+
+            print("Current Git commit is", out.strip())
+
         :param env: A dictionary of environment variables to expose to the
             command. By default, all environment variables are passed.
         :type env: dict or None
         :param bool silent: Silence command output, unless the command fails.
+            If ``True``, returns the command output (unless the command fails).
             ``False`` by default.
         :param success_codes: A list of return codes that are considered
             successful. By default, only ``0`` is considered success.
@@ -258,19 +349,35 @@ class Session:
             ``--error-on-external-run``. This has no effect for sessions that
             do not have a virtualenv.
         :type external: bool
+        :param interrupt_timeout: The timeout (in seconds) that Nox should wait after it
+            and its children receive an interrupt signal before sending a terminate
+            signal to its children. Set to ``None`` to never send a terminate signal.
+            Default: ``0.3``
+        :type interrupt_timeout: float or None
+        :param terminate_timeout: The timeout (in seconds) that Nox should wait after it
+            sends a terminate signal to its children before sending a kill signal to
+            them. Set to ``None`` to never send a kill signal.
+            Default: ``0.2``
+        :type terminate_timeout: float or None
+        :param stdout: Redirect standard output of the command into a file. Can't be
+            combined with *silent*.
+        :type stdout: file or file descriptor
+        :param stderr: Redirect standard output of the command into a file. Can't be
+            combined with *silent*.
+        :type stderr: file or file descriptor
         """
         if not args:
             raise ValueError("At least one argument required to run().")
 
         if self._runner.global_config.install_only:
-            logger.info("Skipping {} run, as --install-only is set.".format(args[0]))
+            logger.info(f"Skipping {args[0]} run, as --install-only is set.")
             return None
 
         return self._run(*args, env=env, **kwargs)
 
     def run_always(
-        self, *args: str, env: Mapping[str, str] = None, **kwargs: Any
-    ) -> Optional[Any]:
+        self, *args: str, env: Mapping[str, str] | None = None, **kwargs: Any
+    ) -> Any | None:
         """Run a command **always**.
 
         This is a variant of :meth:`run` that runs even in the presence of
@@ -298,6 +405,16 @@ class Session:
             ``--error-on-external-run``. This has no effect for sessions that
             do not have a virtualenv.
         :type external: bool
+        :param interrupt_timeout: The timeout (in seconds) that Nox should wait after it
+            and its children receive an interrupt signal before sending a terminate
+            signal to its children. Set to ``None`` to never send a terminate signal.
+            Default: ``0.3``
+        :type interrupt_timeout: float or None
+        :param terminate_timeout: The timeout (in seconds) that Nox should wait after it
+            sends a terminate signal to its children before sending a kill signal to
+            them. Set to ``None`` to never send a kill signal.
+            Default: ``0.2``
+        :type terminate_timeout: float or None
         """
         if (
             self._runner.global_config.no_install
@@ -311,7 +428,9 @@ class Session:
 
         return self._run(*args, env=env, **kwargs)
 
-    def _run(self, *args: str, env: Mapping[str, str] = None, **kwargs: Any) -> Any:
+    def _run(
+        self, *args: str, env: Mapping[str, str] | None = None, **kwargs: Any
+    ) -> Any:
         """Like run(), except that it runs even if --install-only is provided."""
         # Legacy support - run a function given.
         if callable(args[0]):
@@ -340,7 +459,11 @@ class Session:
         return nox.command.run(args, env=env, paths=self.bin_paths, **kwargs)
 
     def conda_install(
-        self, *args: str, auto_offline: bool = True, **kwargs: Any
+        self,
+        *args: str,
+        auto_offline: bool = True,
+        channel: str | Sequence[str] = "",
+        **kwargs: Any,
     ) -> None:
         """Install invokes `conda install`_ to install packages inside of the
         session's environment.
@@ -349,7 +472,7 @@ class Session:
 
             session.conda_install('pandas')
             session.conda_install('numpy', 'scipy')
-            session.conda_install('--channel=conda-forge', 'dask==2.1.0')
+            session.conda_install('dask==2.1.0', channel='conda-forge')
 
         To install packages from a ``requirements.txt`` file::
 
@@ -367,18 +490,23 @@ class Session:
             # Install in editable mode.
             session.install('-e', '.', '--no-deps')
 
+        You can specify a conda channel using `channel=`; a falsey value will
+        not change the current channels. You can specify a list of channels if
+        needed.
+
         Additional keyword args are the same as for :meth:`run`.
 
         .. _conda install:
         """
         venv = self._runner.venv
 
-        prefix_args = ()  # type: Tuple[str, ...]
+        prefix_args: tuple[str, ...] = ()
         if isinstance(venv, CondaEnv):
             prefix_args = ("--prefix", venv.location)
         elif not isinstance(venv, PassthroughEnv):  # pragma: no cover
             raise ValueError(
-                "A session without a conda environment can not install dependencies from conda."
+                "A session without a conda environment can not install dependencies"
+                " from conda."
             )
 
         if not args:
@@ -393,22 +521,29 @@ class Session:
         if "silent" not in kwargs:
             kwargs["silent"] = True
 
-        extraopts = ()  # type: Tuple[str, ...]
+        extraopts: list[str] = []
         if auto_offline and venv.is_offline():
             logger.warning(
-                "Automatically setting the `--offline` flag as conda repo seems unreachable."
+                "Automatically setting the `--offline` flag as conda repo seems"
+                " unreachable."
             )
-            extraopts = ("--offline",)
+            extraopts.append("--offline")
+
+        if channel:
+            if isinstance(channel, str):
+                extraopts.append(f"--channel={channel}")
+            else:
+                extraopts += [f"--channel={c}" for c in channel]
 
         self._run(
-            "conda",
+            venv.conda_cmd,
             "install",
             "--yes",
             *extraopts,
             *prefix_args,
             *args,
             external="error",
-            **kwargs
+            **kwargs,
         )
 
     def install(self, *args: str, **kwargs: Any) -> None:
@@ -434,6 +569,12 @@ class Session:
 
         Additional keyword args are the same as for :meth:`run`.
 
+        .. warning::
+
+            Running ``session.install`` without a virtual environment
+            is no longer supported. If you still want to do that, please
+            use ``session.run("pip", "install", ...)`` instead.
+
         .. _pip: https://pip.readthedocs.org
         """
         venv = self._runner.venv
@@ -444,6 +585,13 @@ class Session:
             raise ValueError(
                 "A session without a virtualenv can not install dependencies."
             )
+        if isinstance(venv, PassthroughEnv):
+            raise ValueError(
+                f"Session {self.name} does not have a virtual environment, so use of"
+                " session.install() is no longer allowed since it would modify the"
+                " global Python environment. If you're really sure that is what you"
+                ' want to do, use session.run("pip", "install", ...) instead.'
+            )
         if not args:
             raise ValueError("At least one argument required to install().")
 
@@ -457,14 +605,28 @@ class Session:
 
     def notify(
         self,
-        target: "Union[str, SessionRunner]",
-        posargs: Optional[Iterable[str]] = None,
+        target: str | SessionRunner,
+        posargs: Iterable[str] | None = None,
     ) -> None:
         """Place the given session at the end of the queue.
 
         This method is idempotent; multiple notifications to the same session
         have no effect.
 
+        A common use case is to notify a code coverage analysis session
+        from a test session::
+
+            @nox.session
+            def test(session):
+                session.run("pytest")
+                session.notify("coverage")
+
+            @nox.session
+            def coverage(session):
+                session.run("coverage")
+
+        Now if you run `nox -s test`, the coverage session will run afterwards.
+
         Args:
             target (Union[str, Callable]): The session to be notified. This
                 may be specified as the appropriate string (same as used for
@@ -482,11 +644,19 @@ class Session:
         """Outputs a log during the session."""
         logger.info(*args, **kwargs)
 
-    def error(self, *args: Any) -> "_typing.NoReturn":
+    def warn(self, *args: Any, **kwargs: Any) -> None:
+        """Outputs a warning during the session."""
+        logger.warning(*args, **kwargs)
+
+    def debug(self, *args: Any, **kwargs: Any) -> None:
+        """Outputs a debug-level message during the session."""
+        logger.debug(*args, **kwargs)
+
+    def error(self, *args: Any) -> NoReturn:
         """Immediately aborts the session and optionally logs an error."""
         raise _SessionQuit(*args)
 
-    def skip(self, *args: Any) -> "_typing.NoReturn":
+    def skip(self, *args: Any) -> NoReturn:
         """Immediately skips the session and optionally logs a warning."""
         raise _SessionSkip(*args)
 
@@ -495,21 +665,21 @@ class SessionRunner:
     def __init__(
         self,
         name: str,
-        signatures: List[str],
+        signatures: list[str],
         func: Func,
         global_config: argparse.Namespace,
-        manifest: "Manifest",
+        manifest: Manifest,
     ) -> None:
         self.name = name
         self.signatures = signatures
         self.func = func
         self.global_config = global_config
         self.manifest = manifest
-        self.venv: Optional[ProcessEnv] = None
-        self.posargs: List[str] = global_config.posargs[:]
+        self.venv: ProcessEnv | None = None
+        self.posargs: list[str] = global_config.posargs[:]
 
     @property
-    def description(self) -> Optional[str]:
+    def description(self) -> str | None:
         doc = self.func.__doc__
         if doc:
             first_line = doc.strip().split("\n")[0]
@@ -518,12 +688,16 @@ class SessionRunner:
 
     def __str__(self) -> str:
         sigs = ", ".join(self.signatures)
-        return "Session(name={}, signatures={})".format(self.name, sigs)
+        return f"Session(name={self.name}, signatures={sigs})"
 
     @property
     def friendly_name(self) -> str:
         return self.signatures[0] if self.signatures else self.name
 
+    @property
+    def tags(self) -> list[str]:
+        return self.func.tags
+
     @property
     def envdir(self) -> str:
         return _normalize_path(self.global_config.envdir, self.friendly_name)
@@ -546,47 +720,44 @@ class SessionRunner:
         if backend is None or backend == "virtualenv":
             self.venv = VirtualEnv(
                 self.envdir,
-                interpreter=self.func.python,  # type: ignore
+                interpreter=self.func.python,  # type: ignore[arg-type]
                 reuse_existing=reuse_existing,
                 venv_params=self.func.venv_params,
             )
-        elif backend == "conda":
+        elif backend in {"conda", "mamba"}:
             self.venv = CondaEnv(
                 self.envdir,
-                interpreter=self.func.python,  # type: ignore
+                interpreter=self.func.python,  # type: ignore[arg-type]
                 reuse_existing=reuse_existing,
                 venv_params=self.func.venv_params,
+                conda_cmd=backend,
             )
         elif backend == "venv":
             self.venv = VirtualEnv(
                 self.envdir,
-                interpreter=self.func.python,  # type: ignore
+                interpreter=self.func.python,  # type: ignore[arg-type]
                 reuse_existing=reuse_existing,
                 venv=True,
                 venv_params=self.func.venv_params,
             )
         else:
             raise ValueError(
-                "Expected venv_backend one of ('virtualenv', 'conda', 'venv'), but got '{}'.".format(
-                    backend
-                )
+                "Expected venv_backend one of ('virtualenv', 'conda', 'mamba',"
+                f" 'venv'), but got '{backend}'."
             )
 
         self.venv.create()
 
-    def execute(self) -> "Result":
-        logger.warning("Running session {}".format(self.friendly_name))
+    def execute(self) -> Result:
+        logger.warning(f"Running session {self.friendly_name}")
 
         try:
-            # By default, nox should quietly change to the directory where
-            # the noxfile.py file is located.
-            cwd = py.path.local(
-                os.path.realpath(os.path.dirname(self.global_config.noxfile))
-            ).as_cwd()
+            cwd = os.path.realpath(os.path.dirname(self.global_config.noxfile))
 
-            with cwd:
+            with _chdir(cwd):
                 self._create_venv()
                 session = Session(self)
+                session.env["NOX_CURRENT_SESSION"] = session.name
                 self.func(session)
 
             # Nothing went wrong; return a success.
@@ -596,6 +767,9 @@ class SessionRunner:
             if self.global_config.error_on_missing_interpreters:
                 return Result(self, Status.FAILED, reason=str(exc))
             else:
+                logger.warning(
+                    "Missing interpreters will error by default on CI systems."
+                )
                 return Result(self, Status.SKIPPED, reason=str(exc))
 
         except _SessionQuit as exc:
@@ -608,13 +782,11 @@ class SessionRunner:
             return Result(self, Status.FAILED)
 
         except KeyboardInterrupt:
-            logger.error("Session {} interrupted.".format(self.friendly_name))
+            logger.error(f"Session {self.friendly_name} interrupted.")
             raise
 
         except Exception as exc:
-            logger.exception(
-                "Session {} raised exception {!r}".format(self.friendly_name, exc)
-            )
+            logger.exception(f"Session {self.friendly_name} raised exception {exc!r}")
             return Result(self, Status.FAILED)
 
 
@@ -622,7 +794,7 @@ class Result:
     """An object representing the result of a session."""
 
     def __init__(
-        self, session: SessionRunner, status: Status, reason: Optional[str] = None
+        self, session: SessionRunner, status: Status, reason: str | None = None
     ) -> None:
         """Initialize the Result object.
 
@@ -651,11 +823,12 @@ class Result:
         """
         if self.status == Status.SUCCESS:
             return "was successful"
+
         status = self.status.name.lower()
         if self.reason:
-            return "{}: {}".format(status, self.reason)
-        else:
-            return status
+            return f"{status}: {self.reason}"
+
+        return status
 
     def log(self, message: str) -> None:
         """Log a message using the appropriate log function.
@@ -672,7 +845,7 @@ class Result:
             log_function = logger.error
         log_function(message)
 
-    def serialize(self) -> Dict[str, Any]:
+    def serialize(self) -> dict[str, Any]:
         """Return a serialized representation of this result.
 
         Returns:
